import codecs
from flask import request
from flask_restful import Resource
from lyrebird import application
from lyrebird.mock import context


class Checker(Resource):
    def get(self, checker_id=None):
        if not checker_id:
            checkers = application.checkers
            script_info_list = [checkers[checker_name].json() for checker_name in checkers]
            return application.make_ok_response(data=script_info_list)

        _checker = application.checkers.get(checker_id)
        if not _checker:
            return context.make_fail_response(f'Checker {checker_id} not found')

        checker = application.checkers[checker_id]
        with codecs.open(checker.path, 'r', 'utf-8') as f:
            content = f.read()
            return application.make_ok_response(data=content)

    def put(self, checker_id):
        _checker = application.checkers.get(checker_id)
        if not _checker:
            return context.make_fail_response(f'Checker {checker_id} not found')

        if 'status' not in request.json:
            return context.make_fail_response(f'Checker {checker_id} status not found')

        _checker_status = request.json.get('status')
        if _checker_status:
            application.checkers[checker_id].activate()
        else:
            application.checkers[checker_id].deactivate()
        return context.make_ok_response()



from flask_restful import Resource
from flask import jsonify, request
from lyrebird import application

# Default event page size
PAGE_SIZE = 20


class Event(Resource):

    def get(self, channel=None, page=0, event_id=None):
        db = application.server['db']
        channel_rules = []
        if channel:
            channel_rules = channel.split('+')
        if event_id:
            page = db.get_page_index_by_event_id(event_id, channel_rules, limit=PAGE_SIZE)
        events = db.get_event(channel_rules, offset=page*PAGE_SIZE, limit=PAGE_SIZE)
        page_count = db.get_page_count(channel_rules, page_size=PAGE_SIZE)
        result = []
        for event in events:
            event_str = event.json()
            result.append(event_str)
        return application.make_ok_response(events=result, page=page, page_count=page_count, page_size=PAGE_SIZE)

    def post(self, channel):
        message = request.json
        event_manager = application.server['event']
        event_manager.publish(channel, message)
        return application.make_ok_response()

class Channel(Resource):

    def get(self):
        db = application.server['db']
        channel_list = db.get_channel_list()
        return jsonify([item[0] for item in channel_list])



from flask_restful import Resource
from lyrebird.mock import context
from flask import jsonify
from lyrebird import version

class Status(Resource):

    def get(self):
        conf = context.application.conf
        return context.make_ok_response(
            **{'ip':conf.get('ip', 'unknown ip'), 
            'mock.port':conf['mock.port'], 
            'proxy.port':conf['proxy.port'],
            'version': version.VERSION
            })

class Manifest(Resource):

    def get(self):
        conf = context.application.conf
        return context.make_ok_response(
            **{'manifest': conf.get('manifest', [])
            })


class WorkMode(Resource):

    def put(self, mode=None):
        if context.Mode.contains(mode):
            context.application.work_mode = mode
            return context.make_ok_response()
        else:
            return context.make_fail_response(f'Unknown mode: {mode}')

    def get(self):
        return jsonify({'mode': context.application.work_mode})



"""
checker example

steps:
- ignore unexepcted object
- prepare useful info
- check data
- update storage data

Info used in channel flow
├── name
├── time
└─┬ flow
  └─┬ request
    ├─┬ headers
    │ └── Content-Type
    ├── method
    └── url

"""


import time
import hashlib
from collections import OrderedDict
from lyrebird import CustomEventReceiver
from urllib.parse import urlparse, urlencode, parse_qsl


# init event receiver instance
event = CustomEventReceiver()

# THRESHOLD_TIME: Range of duplicate requests in seconds
THRESHOLD_TIME = 0.2

# HISTORY_URL: storage history checked url
HISTORY_URL = OrderedDict()

IGNORE_HOSTNAME = [
    'report.meituan.com',
    'frep.meituan.net'
]

IGNORE_PARAMETER = ['token',
                    'version_name',
                    'uuid'
                    ]


@event('flow')
def duplicate_request(msg):

    # 1.ignore unexepcted object
    if ignore_check(msg):
        return

    # 2.prepare useful info
    origin_url = msg['flow']['request']['url']
    sorted_url = sort_params(origin_url[origin_url.rfind('//')+2:])
    request_key_list = [
        sorted_url,
        msg['flow']['request']['method']
    ]
    request_key = get_md5_code(request_key_list)
    request_time = msg['time']

    # 3.check data
    if request_key in HISTORY_URL:

        history_request_time = HISTORY_URL[request_key]['time']
        if request_time - history_request_time < THRESHOLD_TIME:

            url = HISTORY_URL[request_key]['url']

            title = f'Duplicated requests: {url}\n'

            description = f'Duplicated requests: {url}\n'
            description += f'First at {history_request_time}\n'
            description += f'Second at {request_time}\n'

            event.issue(title, description)

    # 4.update storage data
    HISTORY_URL.update({
        request_key: {
            'time': request_time,
            'url': sorted_url
        }
    })

    overdue_urls = []
    for key, value in HISTORY_URL.items():
        if time.time() - value['time'] >= THRESHOLD_TIME:
            overdue_urls.append(key)
        else:
            break
    for overdue_url in overdue_urls:
        del HISTORY_URL[overdue_url]

def ignore_check(msg):
    if msg.get('name') != 'client.request':
        return True
    if urlparse(msg['flow']['request']['url']).hostname in IGNORE_HOSTNAME:
        return True
    return False

def sort_params(url):
    # serialize parameters and remove ignored parameters
    if len(urlparse(url).query):
        origin_parameters = parse_qsl(urlparse(url).query)
        origin_parameters = [param for param in origin_parameters if param[0] not in IGNORE_PARAMETER]
        origin_parameters.sort(key=lambda x:x[0])
        return url.split('?')[0] + '?' + urlencode(origin_parameters)
    else:
        return url

def get_md5_code(keys:list):
    md5_module = hashlib.sha224()
    for key in keys:
        md5_module.update(bytes(key, encoding = "utf8")  )
    return md5_module.hexdigest()



from flask_restful import Resource
from lyrebird.mock.plugin_manager import plugins
from flask import jsonify
from lyrebird.mock import context


class Plugin(Resource):

    def get(self, plugin_name=None):
        plugins_dict = {}
        for name in plugins:
            plugins_dict[name] = dict(plugins[name])

        if not plugin_name:
            return jsonify(plugins_dict)

        _plugin = plugins.get(plugin_name)

        if not _plugin:
            return context.make_fail_response('Not found plugin')
        # /plugin/demo/static/
        return jsonify(plugins_dict[plugin_name])



import codecs
import json
import os
import re
import uuid
from pathlib import Path
from urllib.parse import urlparse

from lyrebird.log import get_logger

PROP_FILE_NAME = '.lyrebird_prop'

logger = get_logger()


class DataManager:

    def __init__(self):
        self.root_path: Path = None
        self.root = {
            "id": str(uuid.uuid4()),
            "name": "$",
            "type": "group",
            "parent_id": None,
            "children": []
        }
        self.id_map = {}
        self.activated_data = {}
        self.activated_group = {}
        self.secondary_activated_data = {}
        self.clipboard = None
        self.save_to_group_id = None
        self.tmp_group = {'id': 'tmp_group', 'type': 'group', 'name': 'tmp-group', 'children': []}

    def set_root(self, root_path):
        """
        Set a new mock data root dir
        -----
        DataManager will reload all data from this new root dir.
        And all activited data will be removed from DataManager.
        """
        _root_path = Path(root_path).expanduser()
        if not _root_path.is_dir():
            raise RootPathIsNotDir(root_path)
        _root_path.mkdir(parents=True, exist_ok=True)
        self.root_path = _root_path
        self.reload()

    def reload(self):
        if not self.root_path:
            raise RootNotSet
        _root_prop_path = self.root_path / PROP_FILE_NAME
        if not _root_prop_path.exists():
            self._save_prop()
        with codecs.open(_root_prop_path) as f:
            _root_prop = json.load(f)
            self.root = _root_prop
            self.id_map = {}
            self._read_node(self.root)
            self._sort_children_by_name()

    def _read_node(self, node):
        if 'id' in node:
            self.id_map[node['id']] = node
        if 'children' in node:
            for child in node['children']:
                self._read_node(child)

    def get(self, _id):
        """
        Get mock group or data by id
        """
        if not self.root:
            raise RootNotSet
        node = self.id_map.get(_id)
        if not node:
            raise IDNotFound(_id)
        if node.get('type') == 'group' or node.get('type') == None:
            return node
        elif node.get('type') == 'data':
            return self._load_data(_id)

    # -----
    # Mock operations
    # -----

    def activate(self, _id):
        """
        Activite data by id
        """
        if not self.root:
            raise RootNotSet
        _node = self.id_map.get(_id)
        if _node:
            self._activate(_node)
        else:
            raise DataNotFound(f'ID:{_id}')
        if _node.get('super_id'):
            _secondary_search_node_id = _node.get('super_id')
            _secondary_search_node = self.id_map.get(_secondary_search_node_id)
            if not _secondary_search_node:
                raise DataNotFound(f'Secondary search node ID: {_secondary_search_node_id}')
            self._activate(_secondary_search_node, secondary_search=True)
        self.activated_group[_id] = _node

    def _activate(self, node, secondary_search=False):
        if node.get('type', '') == 'data':
            _mock_data = self._load_data(node['id'])
            if _mock_data:
                if not secondary_search:
                    self.activated_data[node['id']] = _mock_data
                else:
                    self.secondary_activated_data[node['id']] = _mock_data
        elif node.get('type', '') == 'group':
            if 'children' in node:
                for child in node['children']:
                    self._activate(child, secondary_search=secondary_search)

    def _load_data(self, data_id):
        _data_file = self.root_path / data_id
        if not _data_file.exists():
            raise DataNotFound(f'Data file {_data_file}')
        with codecs.open(_data_file) as f:
            return json.load(f)

    def deactivate(self):
        """
        Clear activated data
        """
        self.activated_data = {}
        self.activated_group = {}
        self.secondary_activated_data = {}

    def get_matched_data(self, flow):
        """
        Find matched mock data from activated data
        """
        _matched_data = []
        for _data_id in self.activated_data:
            _data = self.activated_data[_data_id]
            if self._is_match_rule(flow, _data.get('rule')):
                _matched_data.append(_data)
        if len(_matched_data) <= 0:
            for _data_id in self.secondary_activated_data:
                _data = self.secondary_activated_data[_data_id]
                if self._is_match_rule(flow, _data.get('rule')):
                    _matched_data.append(_data)
        return _matched_data

    def _is_match_rule(self, flow, rules):
        if not rules:
            return False
        for rule_key in rules:
            pattern = rules[rule_key]
            target = self._get_rule_target(rule_key, flow)
            if not re.search(pattern, target):
                return False
        return True

    def _get_rule_target(self, rule_key, flow):
        prop_keys = rule_key.split('.')
        result = flow
        for prop_key in prop_keys:
            result = result.get(prop_key)
            if not result:
                return None
        return result

    # -----
    # Data tree operations
    # -----

    def _get_request_path(self, request):
        path = request.get('path')
        if not path:
            if not request.get('url'):
                return ''
            parsed_url = urlparse(request['url'])
            path = parsed_url.path
        return path

    def add_data(self, parent_id, raw_data):
        if not isinstance(raw_data, dict):
            raise DataObjectSouldBeADict
        if parent_id == 'tmp_group':
            parent_node = self.tmp_group
        else:
            parent_node = self.id_map.get(parent_id)
            if not parent_node:
                raise IDNotFound(parent_id)
            if parent_node['type'] == 'data':
                raise DataObjectCannotContainAnyOtherObject

        data = dict(raw_data)
        data_id = str(uuid.uuid4())
        data['id'] = data_id
        if 'request' in data:
            _data_name = self._get_request_path(data['request'])
            _data_rule = {
                'request.url': f'(?=.*{self._get_request_path(data["request"])})'
            }
            if 'data' in data['request']:
                data['request']['data'] = self._flow_data_2_str(data['request']['data'])
        else:
            _data_name = data.get('name')
            _data_rule = {'request.url': '(?=.*YOUR-REQUEST-PATH)(?=.*PARAMS)'}

        if 'response' in data:
            if 'data' in data['response']:
                data['response']['data'] = self._flow_data_2_str(data['response']['data'])

        data['name'] = _data_name
        data['rule'] = _data_rule

        data_node = {}
        data_node['id'] = data_id
        data_node['name'] = _data_name
        data_node['type'] = 'data'
        data_node['parent_id'] = parent_id

        data_path = self.root_path / data_id
        with codecs.open(data_path, 'w') as f:
            # Save data file
            json.dump(data, f, ensure_ascii=False)
            logger.debug(f'*** Write file {data_path}')
            # Update parent node
            # New data added in the head of child list
            parent_node['children'].insert(0, data_node)
            logger.debug(f'*** Add to node {data_node}')
            # Update ID mapping
            self.id_map[data_id] = data_node
        self._save_prop()
        return data_id

    def _flow_data_2_str(self, data):
        if isinstance(data, str):
            return data
        return json.dumps(data, ensure_ascii=False)

    def add_group(self, parent_id, name):
        if parent_id == None:
            parent_node = self.root
        else:
            parent_node = self.id_map.get(parent_id)
        if not parent_node:
            raise IDNotFound(parent_id)
        if parent_node.get('type') == 'data':
            raise DataObjectCannotContainAnyOtherObject
        # Add group
        group_id = str(uuid.uuid4())
        new_group = {
            'id': group_id,
            'name': name,
            'type': 'group',
            'parent_id': parent_id,
            'children': [],
            'super_id': None
        }
        # New group added in the head of child list
        if 'children' not in parent_node:
            parent_node['children'] = []
        parent_node['children'].insert(0, new_group)
        # Register ID
        self.id_map[group_id] = new_group
        # Save prop
        self._save_prop()
        return group_id

    def delete(self, _id):
        target_node = self.id_map.get(_id)
        if not target_node:
            raise IDNotFound(_id)
        parent_id = target_node.get('parent_id')
        # Remove refs
        if parent_id:
            parent = self.id_map.get(parent_id)
            parent['children'].remove(target_node)
        else:
            self.root['children'].remove(target_node)
        self._delete(_id)
        # Save prop
        self._save_prop()

    def _delete(self, _id):
        target_node = self.id_map.get(_id)
        if not target_node:
            raise IDNotFound(_id)
        if 'children' in target_node and len(target_node['children']) > 0:
            for child in target_node['children']:
                self._delete(child['id'])
        # Delete from ID mapping
        self.id_map.pop(_id)
        # Delete from file system
        if target_node['type'] == 'data':
            data_file_path = self.root_path / _id
            os.remove(data_file_path)

    def cut(self, _id):
        _node = self.id_map.get(_id)
        if not _node:
            raise IDNotFound(_id)
        self.clipboard = {
            'type': 'cut',
            'id': _id,
            'node': _node
        }

    def copy(self, _id):
        _node = self.id_map.get(_id)
        if not _node:
            raise IDNotFound(_id)
        self.clipboard = {
            'type': 'copy',
            'id': _id,
            'node': _node
        }

    def paste(self, parent_id):
        if not self.clipboard:
            raise NoneClipbord
        _parent_node = self.id_map.get(parent_id)
        _node = self.clipboard['node']
        if not _parent_node:
            raise IDNotFound(parent_id)
        if self.clipboard['type'] == 'cut':
            _origin_parent = self.id_map.get(_node['parent_id'])
            _origin_parent['children'].remove(_node)
            _parent_node['children'].insert(0, _node)
            _node['parent_id'] = parent_id
        elif self.clipboard['type'] == 'copy':
            self._copy_node(_parent_node, _node)
        self._save_prop()

    def _copy_node(self, parent_node, node):
        new_node = {}
        new_node.update(node)
        new_node['id'] = str(uuid.uuid4())
        new_node['parent_id'] = parent_node['id']
        # Add to target node
        parent_node['children'].insert(0, new_node)
        # Register ID
        self.id_map[new_node['id']] = new_node
        if new_node['type'] == 'group':
            new_node['children'] = []
            for child in node['children']:
                self._copy_node(new_node, child)
        elif new_node['type'] == 'data':
            self._copy_file(new_node, node)

    def _copy_file(self, target_data_node, data_node):
        _id = data_node['id']
        origin_file_path = self.root_path / _id
        new_file_id = target_data_node['id']
        new_file_path = self.root_path / new_file_id
        with codecs.open(origin_file_path, 'r') as inputfile, codecs.open(new_file_path, 'w') as outputfile:
            origin_text = inputfile.read()
            prop = json.loads(origin_text)
            prop['id'] = new_file_id
            new_prop_text = json.dumps(prop, ensure_ascii=False)
            outputfile.write(new_prop_text)

    def _save_prop(self):
        self._sort_children_by_name()
        prop_str = PropWriter().parse(self.root)
        # Save prop
        prop_file = self.root_path / PROP_FILE_NAME
        with codecs.open(prop_file, 'w') as f:
            f.write(prop_str)

    def _sort_children_by_name(self):
        for node_id in self.id_map:
            node = self.id_map[node_id]
            if 'children' not in node:
                continue
            node['children'] = sorted(node['children'], key=lambda sub_node: sub_node['name'])

    # -----
    # Conflict checker
    # -----

    def check_conflict(self, _id):
        node = self.id_map.get(_id)
        if not node:
            raise IDNotFound(_id)
        data_array = []

        def _read_data(node):
            if node.get('type') == 'data':
                _data_file = self.root_path / node['id']
                with codecs.open(_data_file, 'r') as f:
                    _data = json.load(f)
                    _data['parent_id'] = node['parent_id']
                    data_array.append(_data)
            elif node.get('type') == 'group':
                for child in node['children']:
                    _read_data(child)
        _read_data(node)
        return self.check_conflict_data(data_array)

    def activated_data_check_conflict(self):
        data_array = list(self.activated_data.values())
        return self.check_conflict_data(data_array)

    def check_conflict_data(self, data_array):
        conflict_rules = []
        for _data in data_array:
            _rule = _data['rule']
            _hit_data = []
            for _test_data in data_array:
                if self._is_match_rule(_test_data, _rule):
                    _target_node = {
                        'id': _test_data['id'],
                        'name': _test_data['name'],
                        'url': _test_data['request']['url'],
                        'abs_parent_path': self._get_abs_parent_path(_test_data)
                    }
                    _hit_data.append(_target_node)
            if len(_hit_data) > 1:
                _src_node = {
                    'id': _data['id'],
                    'name': _data['name'],
                    'rule': _data['rule'],
                    'abs_parent_path': self._get_abs_parent_path(_data)
                }
                conflict_rules.append(
                    {
                        'data': _src_node,
                        'conflict_data': _hit_data
                    }
                )
        return conflict_rules

    def _get_abs_parent_path(self, node, path='/'):
        parent_node = self._get_node_parent(node)
        if parent_node is None:
            return path
        current_path = '/' + node['name'] + path
        return self._get_abs_parent_path(parent_node, path=current_path)

    def _get_abs_parent_obj(self, node, parent_obj=None):
        if parent_obj is None:
            parent_obj = []
        if 'id' not in node:
            return parent_obj
        node_info = self.id_map.get(node['id'])
        if not node_info:
            return parent_obj
        parent_obj.insert(0, {
            'id': node_info['id'],
            'name': node_info['name'],
            'type': node_info['type'],
            'parent_id': node_info['parent_id']
        })
        parent_node = self._get_node_parent(node)
        if parent_node is None:
            return parent_obj
        return self._get_abs_parent_obj(parent_node, parent_obj=parent_obj)

    def _get_node_parent(self, node):
        if 'parent_id' not in node:
            return None
        parent_node = self.id_map.get(node['parent_id'])
        if not parent_node:
            return None
        return parent_node

    # -----
    # Record API
    # -----

    def save_data(self, data):
        if len(self.activated_group) > 0:
            # TODO use self.save_to_group_id
            target_group_id = list(self.activated_group.keys())[0]
            self.add_data(target_group_id, data)
        else:
            self.add_data('tmp_group', data)

    # -----
    # Editor
    # -----

    def update_group(self, _id, data):
        ignore_keys = ['id', 'parent_id', 'type', 'children']
        node = self.id_map.get(_id)
        if not node:
            raise IDNotFound(_id)

        update_data = {k: data[k] for k in data if k not in ignore_keys}
        node.update(update_data)

        delete_keys = [k for k in node if k not in data and k not in ignore_keys]
        for key in delete_keys:
            node.pop(key)
        self._save_prop()

    def update_data(self, _id, data):
        node = self.id_map.get(_id)
        if not node:
            raise IDNotFound(_id)
        node['name'] = data['name']

        data_file = self.root_path / _id
        if not data_file.exists():
            raise DataNotFound(_id)
        with codecs.open(data_file, 'w') as f:
            data_str = json.dumps(data, ensure_ascii=False)
            f.write(data_str)
        self._save_prop()
        if self.activated_data.get(_id):
            self.activate(_id)


# -----------------
# Exceptions
# -----------------

class RootNotSet(Exception):
    pass


class RootPathNotExists(Exception):
    pass


class RootPathIsNotDir(Exception):
    pass


class LyrebirdPropNotExists(Exception):
    pass


class DataNotFound(Exception):
    pass


class DataObjectCannotContainAnyOtherObject(Exception):
    pass


class DataObjectSouldBeADict(Exception):
    pass


class IDNotFound(Exception):
    pass


class NoneClipbord(Exception):
    pass


class DumpPropError(Exception):
    pass


class PropWriter:

    def __init__(self):
        self.indent = 0
        self.parsers = {
            'dict': self.dict_parser,
            'list': self.list_parser,
            'int': self.int_parser,
            'str': self.str_parser,
            'NoneType': self.none_parser
        }

    def parse(self, prop):
        prop_type = type(prop)
        parser = self.parsers.get(prop_type.__name__)
        if not parser:
            raise DumpPropError(f'Not support type {prop_type}')
        return parser(prop)

    def dict_parser(self, val):
        dict_str = '{'
        children = None
        for k, v in val.items():
            if k == 'children':
                children = v
                continue
            dict_str += f'"{k}":{self.parse(v)},'
        if children:
            dict_str += self.children_parser(children)
        if dict_str.endswith(','):
            dict_str = dict_str[:-1]
        dict_str += '}'
        return dict_str

    def list_parser(self, val):
        list_str = '['
        for item in val:
            item_str = self.parse(item)
            list_str += item_str + ','
        if list_str.endswith(','):
            list_str = list_str[:-1]
        list_str += ']'
        return list_str

    def int_parser(self, val):
        return f'{val}'

    def str_parser(self, val):
        return json.dumps(val, ensure_ascii=False)

    def none_parser(self, val):
        return "null"

    def children_parser(self, val):
        self.indent += 1
        children_str = '"children":['
        for child in val:
            child_str = self.parse(child)
            children_str += '\n' + '  '*self.indent + child_str + ','
        if children_str.endswith(','):
            children_str = children_str[:-1]
        children_str += ']'
        self.indent -= 1
        return children_str



from lyrebird.mock import context
from lyrebird.log import get_logger
from flask import Response, stream_with_context


logger = get_logger()


class MockHandler:
    """
    根据当前设置数据组的匹配条件,查找对应的mock数据。
    如果没有找到匹配的数据则交由下一个处理器处理。

    """

    def handle(self, handler_context):
        data = context.application.data_manager.get_matched_data(handler_context.flow)
        if len(data) <= 0:
            return
        handler_context.response = self.data2response(data[0])
        activated_groups = context.application.data_manager.activated_group
        activated_group = list(activated_groups.values())[0]
        logger.info(
            f'<Mock> Hit Group:{activated_group.get("name")} - Data:{data[0]["name"]} \nURL: {handler_context.flow["request"]["url"]}\nGroupID:{activated_group["id"]} DataID:{data[0]["id"]}')

    def data2response(self, data):
        response = data['response']
        code = response['code']
        headers = response['headers']
        headers['lyrebird'] = 'mock'
        resp_data = response['data']

        if resp_data:
            if type(resp_data) == str:
                data_len = len(resp_data.encode())
            else:
                data_len = len(resp_data)
            headers['Content-Length'] = data_len
        else:
            # Handle none response data
            resp_data = ''

        def gen():
            yield resp_data
        return Response(stream_with_context(gen()), status=code, headers=headers)



from flask_restful import Resource
from lyrebird.mock import context
from flask import request
from lyrebird import application


class MockGroup(Resource):
    """
    mock数据组
    """

    def get(self, group_id=None):
        if not group_id:
            root = context.application.data_manager.root
            return application.make_ok_response(data=root)

        # 返回数据组中数据列表
        _group = context.application.data_manager.id_map.get(group_id)
        if not _group:
            return context.make_fail_response('Not found group')

        return application.make_ok_response(data=_group)

    def post(self):
        name = request.json.get('name')
        parent_id = request.json.get('parent_id')
        group_id = context.application.data_manager.add_group(parent_id, name)
        return context.make_ok_response(data={'group_id': group_id})

    def put(self):
        group_id = request.json.get('id')
        data = request.json.get('data')
        context.application.data_manager.update_group(group_id, data)
        return context.make_ok_response()

    def delete(self, group_id):
        context.application.data_manager.delete(group_id)
        return context.make_ok_response()


class MockData(Resource):
    """
    mock数据
    """

    def get(self, _id):
        data = context.application.data_manager.get(_id)
        return application.make_ok_response(data=data)

    def put(self):
        data_id = request.json.get('id')
        context.application.data_manager.update_data(data_id, request.json)
        return context.make_ok_response()

    def post(self):
        parent_id = request.json.get('parent_id')
        data = request.json.get('data')
        new_data_id = context.application.data_manager.add_data(parent_id, data)
        return context.make_ok_response(data_id=new_data_id)

    def delete(self, _id):
        context.application.data_manager.delete(_id)
        return context.make_ok_response()


class ActivatedMockGroup(Resource):

    def get(self):
        return context.make_ok_response(
            data=context.application.data_manager.activated_group
        )

    def put(self, group_id=None, action=None):
        if action == 'activate':
            # Only one group could be activated
            context.application.data_manager.deactivate()
            context.application.data_manager.activate(group_id)
        else:
            context.application.data_manager.deactivate()
        return context.make_ok_response()


class MockGroupByName(Resource):

    def put(self):
        group_name = request.json.get('group_name')
        action = request.json.get('action', 'activate')
        for _id in context.application.data_manager.id_map:
            group = context.application.data_manager.id_map.get(_id)
            if group['name'] == group_name:
                if action == 'activate':
                    context.application.data_manager.deactivate()
                    context.application.data_manager.activate(group['id'])
                else:
                    context.application.data_manager.deactivate()
                return context.make_ok_response()
        return context.make_fail_response(f'Group not found. name={group_name}')



import codecs
import json
from pathlib import Path
import os
from lyrebird.log import get_logger
import uuid
from lyrebird.mock.dm import PropWriter

logger = get_logger()

id_map = {}
root = {
    'id': str(uuid.uuid4()),
    'name': '$',
    'type': 'group',
    'parent_id': None,
    'children': []
}
data_root_dir = None


def check_data_version(data_dir):
    prop_file_name = '.lyrebird_prop'
    data_root_path = Path(data_dir).expanduser()
    if not data_root_path.is_dir():
        raise NotDir(f'{data_dir}')
    # empty dir - mock data v2
    file_list = os.listdir(data_root_path)
    if len(file_list) == 0:
        return '1.7.0'
    prop_file = data_root_path / prop_file_name
    # alpha version
    if not prop_file.exists():
        return '0.15.0'
    with codecs.open(prop_file) as f:
        prop_content = f.read()
    # mock data v1
    if len(prop_content) == 0:
        return '1.0.0'
    json.loads(prop_content)
    # mock data v2
    return '1.7.0'


def update(data_dir):
    global root
    global data_root_dir
    data_root_dir = Path(data_dir)
    _backup_dir = _backup(data_dir)
    for data_groups_dir in Path(_backup_dir).iterdir():
        if not data_groups_dir.is_dir():
            continue
        _load_mock_group(data_groups_dir)
    prop_str = PropWriter().parse(root)
    with codecs.open(Path(data_dir) / '.lyrebird_prop', 'w', 'utf-8') as f:
        f.write(prop_str)


def _load_mock_group(group_dir):
    global id_map
    global root
    logger.log(60, f'Load mock data group from {group_dir}')
    group_prop_file = group_dir / '.lyrebird_prop'
    group_prop = {
        'id': None,
        'name': None,
        'parent_id': root['id'],
        'type': 'group',
        'super_id': None,
        'children': []
    }
    with codecs.open(group_prop_file, 'r', 'utf-8') as f:
        _group_prop = json.load(f)
        if 'parent' in _group_prop:
            group_prop['super_id'] = _group_prop['parent']
            del _group_prop['parent']
        group_prop.update(_group_prop)
    for data_dir in group_dir.iterdir():
        if not data_dir.is_dir():
            continue
        data_prop = _load_mock_data(data_dir, group_prop['id'])
        if data_prop:
            group_prop['children'].append(data_prop)
    root['children'].append(group_prop)
    id_map[group_prop['id']] = group_prop


def _load_mock_data(data_dir, parent_id):
    global id_map
    global data_root_dir
    logger.log(60, f'Load data from {data_dir}')
    data_prop_file = data_dir / '.lyrebird_prop'
    if not data_prop_file.exists():
        logger.error(f'Not found prop file in {data_dir}')
        return
    data_prop = {
        'id': None,
        'name': None,
        'parent_id': parent_id,
        'type': 'data'
    }
    with codecs.open(data_prop_file, 'r', 'utf-8') as f:
        _data_prop = json.load(f)
        data_prop['id'] = _data_prop['id']
        data_prop['name'] = _data_prop['name']
        id_map[data_prop['id']] = data_prop
    _save_data_to_file(data_dir, data_root_dir/data_prop['id'])
    return data_prop


def _save_data_to_file(data_dir, dist, _id=None):
    data_root = Path(data_dir)
    _prop = data_root / '.lyrebird_prop'
    _req = data_root / 'request'
    _req_data = data_root / 'request_data'
    _resp = data_root / 'response'
    _resp_data = data_root / 'response_data'
    prop = json.load(codecs.open(_prop))
    if _id:
        prop['id'] = _id
    if _req.exists():
        prop['request'] = json.load(codecs.open(_req))
        if _req_data.exists():
            with codecs.open(_req_data) as f:
                prop['request']['data'] = f.read()
    if _resp.exists():
        prop['response'] = json.load(codecs.open(_resp))
        if _resp_data.exists():
            with codecs.open(_resp_data) as f:
                prop['response']['data'] = f.read()
    with codecs.open(dist, 'w') as f:
        json.dump(prop, f, ensure_ascii=False)


def _backup(data_dir):
    _backup_dir = str(data_dir)+'_bak'
    os.rename(str(data_dir), str(data_dir)+'_bak')
    os.mkdir(data_dir)
    logger.warning(f'\n----------------\nMock data backup was saved to {_backup_dir}\n----------------\n')
    return _backup_dir


class IDNotFound(Exception):
    pass


class NotDir(Exception):
    pass


class PropFileNotFound(Exception):
    pass



from pathlib import Path
import codecs
import json
from packaging import version
from urllib.parse import urlparse
import requests
import subprocess
import time
import shutil
import jinja2
import json
import os
from lyrebird import log as nlog


logger = nlog.get_logger()


config_template = {
    "version": "0.10.5",
    "proxy.filters": [],
    "proxy.port": 4272,
    "mock.port": 9090,
    "mock.data": "{{current_dir}}/mock_data/personal",
    "mock.proxy_headers": {
        "scheme": "MKScheme",
        "host": "MKOriginHost",
        "port": "MKOriginPort"
    }
}


class ConfigManager():
    ROOT = Path('~/.lyrebird').expanduser()
    DEFAULT_FILENAME = 'conf.json'
    BASE_CONFIG = ROOT/DEFAULT_FILENAME

    def __init__(self, conf_path=None):
        self.config = config_template
        self.config_root = self.ROOT
        self.conf_file = self.BASE_CONFIG

        self.update_base_config()
        self.read_config()
        if conf_path:
            self.update_conf_source(conf_path)

    def update_conf_source(self, path):
        input_path: Path = Path(path).expanduser().absolute()
        if input_path.is_dir():
            input_root = input_path
            input_file = input_path / self.DEFAULT_FILENAME
        else:
            input_root = input_path.parent
            input_file = input_path

        if not input_file.exists():
            logger.error(f'Config {input_file} not found!')
        else:
            self.config_root = input_root
            self.conf_file = input_file
            self.read_config()

    def read_config(self):
        template_env = jinja2.Environment(loader=jinja2.FileSystemLoader(str(self.config_root)))
        template = template_env.get_template(self.conf_file.name)
        loaded_config = json.loads(template.render(current_dir=str(self.config_root), download_dir=str(self.ROOT/'downloads')))
        self.config.update(loaded_config)

    def write_config(self):
        self.config_root.mkdir(parents=True, exist_ok=True)
        with codecs.open(self.conf_file, 'w', 'utf-8') as f:
            f.write(json.dumps(self.config, indent=4, ensure_ascii=False))

    def update_base_config(self):
        if self.BASE_CONFIG.exists() and self.BASE_CONFIG.is_file():
            with codecs.open(self.BASE_CONFIG, 'r', 'utf-8') as f:
                base_conf = json.load(f)
                if version.parse(base_conf.get('version', '0.0.0')) < version.parse(config_template.get('version', '0.0.0')):
                    self.write_config()
        else:
            self.write_config()


class ConfigException(Exception):
    pass


resource_template = {
    'uri': None,
    'config': None
}


class Rescource:
    cache_filename = 'resource.json'
    download_dirname = 'downloads'

    def __init__(self, conf_root_path='~/.lyrebird'):
        self.root = Path(conf_root_path).expanduser().absolute()
        self.cache_file = self.root / self.cache_filename
        self.download_dir = self.root / self.download_dirname
        self.cache = None
        # load cache
        self.load()

    def load(self):
        if self.cache_file.exists():
            with codecs.open(self.cache_file, 'r', 'utf-8') as f:
                try:
                    self.cache = json.load(f)
                except Exception:
                    self.cache = resource_template
        else:
            self.cache = resource_template

    def save(self):
        with codecs.open(self.cache_file, 'w', 'utf-8') as f:
            f.write(json.dumps(self.cache, ensure_ascii=False, indent=4))
            f.close()

    def download(self, uri):
        self.cache['uri'] = uri
        self.save()

        if self.download_dir.exists():
            shutil.rmtree(self.download_dir.absolute())
        self.download_dir.mkdir(exist_ok=True)

        uri = urlparse(self.cache.get('uri'))
        if uri.scheme == 'http' or uri.scheme == 'https':
            self._http()
        elif uri.scheme.startswith('git+'):
            self._git()
        else:
            raise RescourceException(f'Unknown scheme {self.cache}')

    def _git(self):
        git_url = self.cache.get('uri')[4:]
        p = subprocess.run(f'git clone {git_url} {self.download_dir.absolute()}', shell=True)
        p.check_returncode()
        logger.warning(f'Source downloaded to {str(self.download_dir.absolute())}')

    def _http(self):
        # resp = requests.get(self.cache.get('uri'), allow_redirects=True)
        # TODO support http download
        # 1. download gzip file and unzip it
        # 2. download from git repo
        pass


class RescourceException(Exception):
    pass



import requests
from flask import Response, stream_with_context
from requests.packages import urllib3
from .. import context
import urllib
from lyrebird import application
from lyrebird.log import get_logger


# 关闭ssl警告
urllib3.disable_warnings()

logger = get_logger()


class ProxyHandler:
    """
    当前处理链上没有生成response的请求，尝试按照代理规则代理。

    """

    def handle(self, handler_context):
        if handler_context.response:
            handler_context.response.headers.add_header("isMocked", "True")
            return
        request = handler_context.request

        origin_url = handler_context.get_origin_url()

        logger.info(f'<Proxy> {origin_url}')

        if not origin_url:
            return

        parsed_url = urllib.parse.urlparse(origin_url)
        if not parsed_url.hostname:
            return
        elif parsed_url.hostname in ['localhost', '127.0.0.1', ] and parsed_url.port == application.config["mock.port"]:
            handler_context.response = Response(response='Duplicate request path\n', status=400)
            return

        method = request.method
        data = request.get_data() or request.form or None
        headers = dict()
        for name, value in request.headers:
            if not value or name in ['Cache-Control', 'Host']:
                continue
            headers[name] = value

        handler_context.update_server_req_time()
        r = requests.request(method, origin_url, headers=headers, data=data, cookies=request.cookies, stream=True,
                             verify=False)

        # 增加数据源标记，此数据经代理得到
        resp_headers = [('lyrebird', 'proxy')]
        for name, value in r.headers.items():
            # rm 'content-length' from ignore list
            if name.lower() in ('content-encoding',
                                'transfer-encoding'):
                continue
            if name.lower() == 'content-length' and 'content-encoding' in r.headers and r.headers['content-encoding'] == 'gzip':
                # 如果是gzip请求，由于requests自动解压gzip，所以此处抹去content-length,以匹配解压后的数据长度
                continue
            resp_headers.append((name, value))

        handler_context.request.url = origin_url

        # After huangyuanzhen test, we use 2048byte buffer :D
        handler_context.response = Response(
            stream_with_context(r.iter_content(chunk_size=2048)),
            status=r.status_code,
            headers=resp_headers)



import os
from flask import render_template, Blueprint, send_file
import traceback
from lyrebird.mock import context
from lyrebird import log
from ... import version
from .. import plugin_manager
from bs4 import BeautifulSoup
import datetime
from pathlib import Path
from lyrebird import reporter


CLIENT_ROOT_DIR = Path(__file__).parent/'../../client/static'


ui = Blueprint('ui', __name__, url_prefix='/ui', static_folder=str(CLIENT_ROOT_DIR))


logger = log.get_logger()


def render_with_plugin(template_name_or_list, **context):
    web_plugins = []
    icon = ''
    for web_plugin_name in plugin_manager.web_plugins:
        try:
            obj = plugin_manager.web_plugins[web_plugin_name]
            if hasattr(obj, "index") and not obj.index():
                continue
            if hasattr(obj, "get_icon"):
                icon = obj.get_icon()
            if hasattr(obj, 'get_title'):
                title = obj.get_title()
            else:
                title = web_plugin_name.capitalize()
            web_plugins.append({'name': web_plugin_name, 'icon': icon, 'title': title})
        except Exception:
            logger.error(f'Load web plugin error. {traceback.format_exc()}')
    return render_template(template_name_or_list, web_plugins=web_plugins, version=version.VERSION, **context)


@ui.route('/')
def index():
    reporter.page_in('inspector')
    return send_file(str(CLIENT_ROOT_DIR/'index.html'))



import pkg_resources
from lyrebird.mock import handlers
from collections import OrderedDict, namedtuple
import traceback
from jinja2 import Environment, PackageLoader
from flask import send_from_directory
import sys
import os
import inspect
import codecs
import json
from types import FunctionType
from pathlib import Path
from lyrebird.mock import context
from lyrebird import log
from flask import Blueprint
from .plugin_view import PluginView


"""
Plugin manager
"""

logger = log.get_logger()


ROOT_DIR = Path('~/.lyrebird').expanduser()
PLUGINS_DIR = ROOT_DIR/'plugins'


DATA_HANDLER_ENTRY_POINT = 'lyrebird_data_handler'
WEB_ENTRY_POINT = 'lyrebird_web'

# v1.0 plugin entry point
PLUGIN_ENTRY_POINT = 'lyrebird_plugin'

inner_handler = OrderedDict()
data_handler_plugins = OrderedDict()
web_plugins = OrderedDict()

# Store all plugins with project name
# Contains v1.0 plugins and older plugins
plugins = {}

# plugin blueprint list for add rule
plugin_blueprint_list = []


def _ep_to_plugin(ep, **kwargs):
    plugin = dict(
        name=ep.name, 
        version=ep.dist.version, 
        project_name=ep.dist.project_name, 
        **kwargs)
    plugin_key = plugin['project_name']
    target_plugin = plugins.get(plugin_key)
    if target_plugin:
        target_plugin.update(plugin)
    else:
        plugins[plugin_key] = plugin


def _plugin_entry_check(ep_entry):
    pass


def _load_plugin():
    for ep in pkg_resources.iter_entry_points(PLUGIN_ENTRY_POINT):
        plugin_entry = ep.load()
        if not _plugin_entry_check(plugin_entry).result:
            logger.error(f'Load plugin {plugin_entry.name} error . skip')
            continue
        

def _load_data_handler_plugin():
    data_handler_plugins.clear()
    for ep in pkg_resources.iter_entry_points(DATA_HANDLER_ENTRY_POINT):
        try:
            handler_class = ep.load()
            if not isinstance(handler_class, type):
                logger.error(f'Load plugin {ep} error. Entry point is not a class')
                continue
            if not hasattr(handler_class, 'handle'):
                logger.error(f'Load plugin {ep} error. Not found function - "handle"')
                continue
            handler = handler_class()
            data_handler_plugins[ep.name] = handler
            # Add plugin handler to new plugin object
            _ep_to_plugin(ep, beta_handler=handler)
        except Exception:
            logger.error(f'Load plugin {ep} error')
            traceback.print_exc()


def _load_web_plugin():
    web_plugins.clear()
    for ep in pkg_resources.iter_entry_points(WEB_ENTRY_POINT):
        try:
            web_class = ep.load()
        except Exception:
            logger.error(f'Load plugin fail. {ep}')
            traceback.print_exc()
            continue
        if not isinstance(web_class, type):
            logger.error(f'Load plugin {ep} error. Entry point is not a class')
            continue
        web_plugin = web_class()
        if ep.name in web_plugins:
            logger.error(f'Plugin {ep} duplicate name with {web_plugins[ep.name]}')
            continue
        if not hasattr(web_class, 'on_create'):
            logger.error(f'Load plugin {ep} error. Not found function - "on_create"')
            continue
        if not hasattr(web_class, 'index'):
            logger.error(f'Load plugin {ep} error. Not found function - "index"')
            continue
        try:
            web_plugin.on_create()
            web_plugin.after_on_create()
        except Exception:
            logger.error(f'Plugin {ep} on create', traceback.format_exc())
            continue
        web_plugins[ep.name] = web_plugin
        # Add web plugin to new plugin object
        _ep_to_plugin(ep, beta_web=web_plugin)

def _load_inner_plugin():
    inner_handler.update(handlers.get_inner_handlers())


def load():
    _load_inner_plugin()
    _load_data_handler_plugin()
    _load_web_plugin()


def add_plugin_to_blueprint(plugin, blueprint):
    """
    NEW add rule to blueprint
    """
    if not plugin:
        return
    try:
        for rule in plugin.rules:
            # 去掉自定义路由前面的"/"
            if rule.rule.startswith('/'):
                rule.rule = rule.rule[1:]
            # 为了防止重名，拼接endpoint
            if not rule.endpoint:
                rule.endpoint = plugin.name + '_' + rule.view_func.__name__
            # 注册API
            blueprint.add_url_rule(f'/{rule.rule}', rule.endpoint, view_func=rule.view_func, **rule.options)
    except Exception:
        logger.error(f'Can not add rule {rule}')
        traceback.print_exc()


def add_view_to_blueprint(blueprint):
    if len(web_plugins) == 0:
        return
    for plugin_name in web_plugins:
        plugin = web_plugins[plugin_name]
        try:
            for rule in plugin.rules:
                # 去掉自定义路由前面的"/"
                if rule.rule.startswith('/'):
                    rule.rule = rule.rule[1:]
                # 为了防止重名，拼接endpoint
                if not rule.endpoint:
                    rule.endpoint = plugin_name + '_' + rule.view_func.__name__
                # 注册视图
                blueprint.add_url_rule(
                    f'/plugin/{plugin_name}/{rule.rule}',
                    rule.endpoint,
                    view_func=rule.view_func,
                    **rule.options
                )
        except Exception:
            logger.error(f'Can not add rule {rule}')
            traceback.print_exc()


def add_event_rules(socket_io):
    if len(web_plugins) == 0:
        return
    for plugin_name in web_plugins:
        plugin = web_plugins[plugin_name]
        try:
            for event_rule in plugin.event_rules:
                socket_io.on_event(event_rule.event, event_rule.handler, event_rule.namespace)
        except Exception:
            logger.error(f'Can not add event rule {event_rule}')
            traceback.print_exc()


def caller_info(index=1):
    stack = inspect.stack()
    caller_module = inspect.getmodule(stack[index].frame)
    caller_module_name = caller_module.__name__
    if caller_module_name.find('.') > 0:
        caller_top_module_name = caller_module_name.split('.')[0]
    else:
        caller_top_module_name = caller_module_name
    CallerInfo = namedtuple('CallerInfo', 'top_module_name module_name')
    return CallerInfo(module_name=caller_module_name, top_module_name=caller_top_module_name)


class PluginNotFoundError(Exception):
    pass


def _get_plugin_conf_path(name):
    plugin = web_plugins.get(name)
    if not plugin:
        raise PluginNotFoundError('Plugin %s not fonud' % name)
    if '.' in plugin.__module__:
        top_module_name = plugin.__module__.split('.')[0]
    else:
        top_module_name = plugin.__module__
    plugin_storage_dir = PLUGINS_DIR/top_module_name
    if not plugin_storage_dir.exists():
        plugin_storage_dir.mkdir()
    return plugin_storage_dir/'conf.json'


def get_conf(name):
    conf_path = _get_plugin_conf_path(name)
    if not conf_path.exists():
        set_default_conf(name)
    with codecs.open(conf_path, 'r', 'utf-8') as f:
        return json.load(f)


def set_conf(name, conf):
    conf_path = _get_plugin_conf_path(name)
    if not conf_path.exists():
        conf_path.touch()
    with codecs.open(conf_path, 'w', 'utf-8') as f:
        f.write(json.dumps(conf, ensure_ascii=False, indent=4))
    context.application.event_bus.publish('config', dict(name=name, action='update'))


def set_default_conf(name):
    plugin = web_plugins[name]
    set_conf(name, plugin.default_conf())
    context.application.event_bus.publish('config', dict(name=name, action='update'))



from flask import Blueprint, request
from flask_restful import Resource, Api
from .common import Status, WorkMode, Manifest
from .flow import Flow, FlowList
from .mock import MockData, MockGroup, ActivatedMockGroup, MockGroupByName
from .config import Conf, ResetConf
from .plugin import Plugin
from .menu import Menu
from .notice import Notice
from .checker import Checker
from .event import Event, Channel
from .conflict_check import ConflictCheck, ActivatedDataConflictCheck
from .mock_editor import Cut, Copy, Paste
from .search import SearchMockDataByName
from lyrebird.log import get_logger
from lyrebird import application
from flask import got_request_exception
import traceback


logger = get_logger()


api = Blueprint('api', __name__, url_prefix='/api')
api_source = Api(api, errors=Exception)


@api.after_request
def after_request(response):
    """
    输出每条请求概要信息
    """
    lyrebird_info = response.headers.get('lyrebird', default='')
    logger.debug(f'[On API]{response.status_code} {lyrebird_info} {request.method} {request.url[:100]}')
    return response


@api.errorhandler(Exception)
def on_500(error):
    return application.make_fail_response(f'{error.__class__.__name__} {error}')


api_source.add_resource(Status, '/status')
api_source.add_resource(Manifest, '/manifest')
api_source.add_resource(Flow, '/flow/<string:id>')
api_source.add_resource(FlowList, '/flow')
api_source.add_resource(Cut, '/cut/<string:_id>')
api_source.add_resource(Copy, '/copy/<string:_id>')
api_source.add_resource(Paste, '/paste/<string:_id>')
api_source.add_resource(MockGroup, '/group', '/group/<string:group_id>')
api_source.add_resource(MockData, '/data', '/data/<string:_id>')
api_source.add_resource(SearchMockDataByName, '/search/group/name/', '/search/group/name/<string:search_str>')
api_source.add_resource(ConflictCheck, '/conflict/id/<string:group_id>')
api_source.add_resource(ActivatedDataConflictCheck, '/conflict/activated')
api_source.add_resource(ActivatedMockGroup, '/mock/activated', '/mock/<string:group_id>/<string:action>')
api_source.add_resource(MockGroupByName, '/mock_by_name')
api_source.add_resource(Conf, '/conf/<string:plugin_name>')
api_source.add_resource(ResetConf, '/conf/<string:plugin_name>/reset')
api_source.add_resource(WorkMode, '/mode', '/mode/<string:mode>')
api_source.add_resource(Plugin, '/plugin', '/plugin/<string:plugin_name>')
api_source.add_resource(Menu, '/menu')
api_source.add_resource(Notice, '/notice')
api_source.add_resource(Checker, '/checker', '/checker/<string:checker_id>')
api_source.add_resource(
    Event,
    '/event',
    '/event/page/<int:page>',
    '/event/id/<string:event_id>',
    '/event/<string:channel>',
    '/event/<string:channel>/page/<int:page>',
    '/event/<string:channel>/id/<string:event_id>'
)
api_source.add_resource(Channel, '/channel')



import math


def convert_size(size_bytes):
   if size_bytes == 0:
       return "0B"
   size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
   i = int(math.floor(math.log(size_bytes, 1024)))
   p = math.pow(1024, i)
   s = round(size_bytes / p, 2)
   return "%s %s" % (s, size_name[i])


def convert_time(duration):
    if duration < 1:
        return str(round(duration * 1000)) + 'ms'
    else:
        return str(round(duration, 2)) + 's'



from flask_restful import Resource
from lyrebird.mock import context
from lyrebird import application


class Cut(Resource):

    def put(self, _id):
        context.application.data_manager.cut(_id)
        return application.make_ok_response()


class Copy(Resource):

    def put(self, _id):
        context.application.data_manager.copy(_id)
        return application.make_ok_response()


class Paste(Resource):

    def put(self, _id):
        context.application.data_manager.paste(_id)
        return application.make_ok_response()



import runpy
import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

VERSION = runpy.run_path(
    os.path.join(here, "lyrebird", "version.py")
)["VERSION"]

with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='lyrebird',
    version=VERSION,
    packages=['lyrebird'],
    url='https://github.com/meituan/lyrebird',
    author='HBQA',
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
    ],
    entry_points={
        'console_scripts': [
            'lyrebird = lyrebird.manager:main'
        ]
    },
    install_requires=[
        "beautifulsoup4==4.7.1",
        "colorama==0.4.1",
        "elasticsearch==7.0.1",
        "Flask==1.1.1",
        "Flask-RESTful==0.3.7",
        "Flask-SocketIO==4.2.1",
        "mitmproxy==4.0.4",
        "packaging==19.0",
        "portpicker==1.3.1",
        "requests==2.21.0",
        "SQLAlchemy==1.3.1",
        "click==6.7",
        "urllib3==1.24.1"
    ],
    extras_require={
        'dev': [
            "autopep8",
            "pylint",
            "pytest"
        ]
    }
)



from .. import context
from lyrebird import application
from lyrebird.log import get_logger
from lyrebird import utils
from urllib.parse import urlparse, unquote
import urllib
import uuid
import time
import gzip
import json
import ipaddress
import binascii


logger = get_logger()


class HandlerContext:
    """
    请求处理器上下文变量
    用于保存一个请求处理过程中的request, response

    """
    MOCK_PATH_PREFIX = '/mock'

    def __init__(self, request):
        self.id = str(uuid.uuid4())
        self.request = request
        self._response = None
        self.client_req_time = None
        self.client_resp_time = None
        self.server_req_time = None
        self.server_resp_time = None
        self.flow = dict(
            id=self.id,
            size=0,
            duration=0,
            start_time=time.time())
        self.client_address = None
        self._parse_request()

    def _parse_request(self):
        # Read stream
        self.request.get_data()
        # parse path
        request_info = self._read_origin_request_info_from_url()
        if not request_info['host']:
            request_info_from_header = self._read_origin_request_info_from_header()
            if len(request_info_from_header) > 0:
                request_info = request_info_from_header

        headers = {k: v for k, v in self.request.headers}
        _request = dict(
            headers=headers,
            method=self.request.method,
            query=self.request.args,
            timestamp=round(time.time(), 3)
        )
        _request.update(request_info)

        # handle request data
        if self.request.method in ['POST', 'PUT']:
            RequestDataHelper.req2dict(self.request, output=_request)

        if self.request.headers.get('Lyrebird-Client-Address'):
            self.client_address = self.request.headers.get('Lyrebird-Client-Address')
        else:
            self.client_address = self.request.remote_addr
        self.flow['client_address'] = self.client_address

        self.flow['request'] = _request
        context.application.cache.add(self.flow)

        logger.debug(f'[On client request] {self.flow["request"]["url"]}')

    def _read_origin_request_info_from_url(self):
        url_prefix = '/'+self.request.blueprint+'/'
        raw_url = self.request.path[len(url_prefix):]
        if self.request.query_string:
            raw_url += '?' + self.request.query_string.decode()
        parsed_path = urlparse(raw_url)
        # urllib.unquote : fix bug - url contains ',' will be auto encoded by flask, that cause proxy not work.
        # e.g /1.2,3 -> 1.2%2C3
        _request = dict(
            url=raw_url,
            scheme=parsed_path.scheme,
            host=parsed_path.hostname,
            port=parsed_path.port if parsed_path.port else '80',
            path=unquote(parsed_path.path)
        )
        return _request

    def _read_origin_request_info_from_header(self):
        proxy_headers = application.config['mock.proxy_headers']
        scheme = self.request.headers.get(proxy_headers['scheme'], default='http')
        host = self.request.headers.get(proxy_headers['host'])
        port = self.request.headers.get(proxy_headers['port'], default='80')
        if not host:
            return {}
        scheme = scheme.strip()
        host = host.strip()
        port = port.strip()
        # if host is IP address then full_host=host:port
        # else if is a domain the full_host=host
        full_host = host
        try:
            ipaddress.ip_address(host)
            full_host = host + ':' + port
        except Exception:
            pass
        return dict(
            url=scheme+'://'+full_host+self.request.full_path[len(self.MOCK_PATH_PREFIX):],
            scheme=scheme,
            host=host,
            port=port,
            path=self.request.path[len(self.MOCK_PATH_PREFIX):]
        )

    @property
    def response(self):
        return self._response

    @response.setter
    def response(self, val):
        self._response = val
        self.update_server_resp_time()

        _response = dict(
            code=self._response.status_code,
            headers={k: v for (k, v) in self._response.headers},
            timestamp=round(time.time(), 3)
        )

        ResponseDataHelper.resp2dict(self._response, output=_response)
        self.flow['response'] = _response

        if val.content_length:
            self.flow['size'] = val.content_length
        else:
            self.flow['size'] = len(val.data)
        self.flow['duration'] = self.server_resp_time - self.client_req_time

        if context.application.work_mode == context.Mode.RECORD:
            dm = context.application.data_manager
            dm.save_data(self.flow)

    def _read_response_info(self):
        self._response.headers.get('Content-Type')

    def update_client_req_time(self):
        self.client_req_time = time.time()
        # 消息总线 客户端请求事件，启用此事件
        method = self.flow['request']['method']
        url = self.flow['request']['url']
        context.application.event_bus.publish('flow.request',
        dict(
            flow=self.flow,
            message=f"URL: {url}\nMethod: {method}\n"
            )
        )

    def update_client_resp_time(self):
        self.client_resp_time = time.time()
        # 消息总线 客户端响应事件，启用此事件
        method = self.flow['request']['method']
        url = self.flow['request']['url']
        code = self.flow['response']['code']
        duration = utils.convert_time(self.flow['duration'])
        size = utils.convert_size(self.flow['size'])
        context.application.event_bus.publish('flow',
                                              dict(
                                                  flow=self.flow,
                                                  message=f"URL: {url}\nMethod: {method}\nStatusCode: {code}\nDuration: {duration}\nSize: {size}"
                                              )
                                              )

    def update_server_req_time(self):
        self.server_req_time = time.time()
        # 消息总线 向服务端转发请求事件，暂不使用
        # context.application.event_bus.publish('flow',
        #                                       dict(name='server.request',
        #                                       time=self.server_req_time,
        #                                       id=self.id,
        #                                       flow=self.flow))

    def update_server_resp_time(self):
        self.server_resp_time = time.time()
        # 消息总线 服务端响应请求事件，暂不使用
        # context.application.event_bus.publish('flow',
        #                                       dict(name='server.response',
        #                                       time=self.server_resp_time,
        #                                       id=self.id,
        #                                       flow=self.flow))

    def get_origin_url(self):
        return self.flow['request'].get('url')


class DataHelper:

    @staticmethod
    def data2Str(data):
        try:
            return data.decode('utf-8')
        except Exception as e:
            logger.warning(f'Data to string failed. {e}')
            return binascii.b2a_base64(data).decode('utf-8')


class RequestDataHelper(DataHelper):

    @staticmethod
    def req2dict(request, output=None):
        if not output:
            output = {}
        content_encoding = request.headers.get('Content-Encoding')
        # Content-Encoding handler
        unziped_data = None

        try:
            if content_encoding and content_encoding == 'gzip':
                unziped_data = gzip.decompress(request.data)

            content_type = request.headers.get('Content-Type')
            if not content_type:
                output['data'] = RequestDataHelper.data2Str(request.data)
                return

            content_type = content_type.strip()
            if content_type.startswith('application/x-www-form-urlencoded'):
                if unziped_data:
                    output['data'] = urllib.parse.parse_qs(unziped_data.decode('utf-8'))
                else:
                    output['data'] = request.form.to_dict()
            elif content_type.startswith('application/json'):
                if unziped_data:
                    output['data'] = json.loads(unziped_data.decode('utf-8'))
                else:
                    output['data'] = request.json
            elif content_type.startswith('text/xml'):
                if unziped_data:
                    output['data'] = unziped_data.decode('utf-8')
                else:
                    output['data'] = request.data.decode('utf-8')
            else:
                # TODO write bin data
                output['data'] = RequestDataHelper.data2Str(request.data)
        except Exception as e:
            output['data'] = RequestDataHelper.data2Str(request.data)
            logger.warning(f'Convert request data fail. {e}')


class ResponseDataHelper(DataHelper):

    @staticmethod
    def resp2dict(response, output=None):
        if not output:
            output = {}
        content_type = response.headers.get('Content-Type')
        if not content_type:
            output['binary_data'] = 'bin'
        else:
            content_type = content_type.strip()

        try:
            if content_type.startswith('application/json'):
                output['data'] = response.json
            elif content_type.startswith('text/xml'):
                output['data'] = response.data.decode('utf-8')
            elif content_type.startswith('text/html'):
                output['data'] = response.data.decode('utf-8')
            else:
                # TODO write bin data
                output['data'] = ResponseDataHelper.data2Str(response.data)
        except Exception as e:
            output['data'] = ResponseDataHelper.data2Str(response.data)
            logger.warning(f'Convert response failed. {e}')



from flask_restful import Resource
from lyrebird.mock import context
from flask import request, jsonify, abort
from lyrebird import application


class Notice(Resource):

    def get(self):
        return context.make_ok_response(
            noticeList=application.notice.notice_list,
            notRemindList=application.notice.not_remind_list
        )

    def put(self):
        notice_status = request.json.get('status')
        unique_key = request.json.get('key')
        application.notice.change_notice_status(unique_key, notice_status)
        return context.make_ok_response()

    def delete(self):
        unique_key = request.json.get('key')
        application.notice.delete_notice(unique_key)
        return context.make_ok_response()
